/***************************************************************************
                          kgesture.cpp  -  description
                             -------------------
    begin                : Wed May 16 11:06:41 EDT 2001
    copyright            : (C) 2001 by Mike Pilone
    email                : mpilone@slac.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

#include <qtimer.h>
#include <qpainter.h>
#include <qpixmap.h>

#include <kdebug.h>
#include <kapp.h>
#include <kiconloader.h>
#include <kglobal.h>
#include <klocale.h>
#include <kpopupmenu.h>
#include <kprocess.h>
#include <kconfig.h>

#include "kgesture.h"
#include "gesture.h"

#define POLL_TIME 5
#define MOUSE_SLEEP_PERIOD 30

#define MENU_ENABLED_ID 1

KGesture::KGesture()
  : KSystemTray(), _display(NULL), _xCoord(0), _yCoord(0),
    _lastGridCoord(0, 0), _mousePaused(0),
    _numStrokesRecorded(0), _lifeLength(3),
    _recognizer(this, "recognizer"),
    _configDialog(this, "configDialog"),
    _recognizedDialog(this, "recognizedDialog"),
    _mousePollTimer(this, "mousePollTimer"),
    _changeIconTimer(this, "changeIconTimer"),
    _showRecognizedDialog(true), _systemBeep(false)
  {
  // set the icon
  setPixmap(KGlobal::iconLoader()->loadIcon("kgesture",
					    KIcon::Small));
  
  // Connect to the X server - needed for mouse events
  initXCon();

  // Build the context menu
  initMenu();
  
  connect(&_mousePollTimer, SIGNAL(timeout()),
	  this, SLOT(slotMousePollTimeout()));
  connect(&_changeIconTimer, SIGNAL(timeout()),
	  this, SLOT(slotChangeIconTimeout()));
  
  readOptions(*(kapp->config()));
  }

KGesture::~KGesture()
  {
  writeOptions(*(kapp->config()));
  }

void KGesture::initXCon()
  {
  _display = KApplication::kApplication()->getDisplay();

  _root = RootWindow(_display,DefaultScreen(_display));
  }

void KGesture::initMenu()
  {
  KPopupMenu *menu = contextMenu();
  menu->setCheckable(true);
  menu->insertItem(i18n("Enabled"), this, SLOT(slotMenuEnable()),
		   0, MENU_ENABLED_ID);
  menu->insertItem(i18n("Preferences..."), this, SLOT(slotMenuPreferences()));
  }

void KGesture::gestureRecognized(Gesture &gest)
  {
  // Try to run the action
  QString action = gest.getAction();

  if (_showRecognizedDialog)
    {
    _recognizedDialog.show(gest);
    }

  // Change the icon so if they can't see the dialog, they know it
  // got the gesture
  setPixmap(KGlobal::iconLoader()->loadIcon("kgesturealert",
					    KIcon::Small));
  _changeIconTimer.start(_lifeLength * 1000, true);

  // If they wanna beep, let'em have it
  if (_systemBeep)
    kapp->beep();
  
  if ((action != QString::null) && (action.length() > 0))
    {
    // Run it here
    KProcess process;
    Q_UINT32 start = 0;
    int length = 0;
    
    while ((length = action.find(" ", start)) != -1)
      {
      length = length-start;
      //kdDebug() << action.mid(start, length) << "\n";
      process << action.mid(start, length);
      
      start += length + 1;
      }
    
    process << action.mid(start, action.length()-start);
    process.start(KProcess::DontCare, KProcess::NoCommunication);
    }
  }

void KGesture::readOptions(KConfig &config)
  {
  config.setGroup("General");
  if (config.readBoolEntry("Enabled", true))
    setEnabled(true);

  config.setGroup("Feedback");
  _showRecognizedDialog = config.readBoolEntry("Display Recognized", true);
  _systemBeep = config.readBoolEntry("System Beep", false);
  _lifeLength = config.readNumEntry("Life Length", 3);
  
  _recognizer.readOptions(config);
  _recognizedDialog.readOptions(config);
  }

void KGesture::writeOptions(KConfig &config)
  {
  _recognizer.writeOptions(config);
  _recognizedDialog.writeOptions(config);
  }

void KGesture::setEnabled(bool state)
  {
  KPopupMenu *menu = contextMenu();

  if (!state)
    {
    menu->setItemChecked(MENU_ENABLED_ID, false);
    _mousePollTimer.stop();
    stroke_trans(_sequence);
    }
  else
    {
    menu->setItemChecked(MENU_ENABLED_ID, true);
    _mousePollTimer.start(POLL_TIME);
    _mousePaused = 0;
    }

  kapp->config()->setGroup("General");
  kapp->config()->writeEntry("Enabled",
			     contextMenu()->isItemChecked(MENU_ENABLED_ID));
  }


/////////////////////////////
// SLOTS

void KGesture::slotMousePollTimeout()
  {
  XQueryPointer (_display, _root, &_rootIDRet, &_childIDRet, &_xCoord,
		 &_yCoord, &_winX, &_winY, &_stateMask);

  QPoint gridCoord(_xCoord, _yCoord);

  //kdDebug() << "Num points recorded: " << _gridPointList.count() << "\n";
  
  if (_lastGridCoord == gridCoord)
    {
    if ((_mousePaused > MOUSE_SLEEP_PERIOD) && (_numStrokesRecorded > 1))
      {
      // process list
      if (stroke_trans (_sequence) == TRUE)
	{
        Gesture *gest = _recognizer.lookupGesture(Gesture(_sequence));
	if (gest)
	  gestureRecognized(*gest);
	}

      // Add this point for the next translation
      stroke_record(gridCoord.x(), gridCoord.y());
      _numStrokesRecorded = 1;
      }

    _mousePaused++;

    // sanity check
    if (_mousePaused > 10000)
      _mousePaused = MOUSE_SLEEP_PERIOD+1;
    }
  else
    {
    _mousePaused = 0;

    stroke_record(gridCoord.x(), gridCoord.y());
    _numStrokesRecorded++;
    }

  _lastGridCoord = gridCoord;
  }

void KGesture::slotMenuEnable()
  {
  setEnabled(!contextMenu()->isItemChecked(MENU_ENABLED_ID));
  }

void KGesture::slotMenuPreferences()
  {
  // Stop the timer, we won't get events while configuring
  _mousePollTimer.stop();

  // Clear the stroke list so the config dialog can use the library
  stroke_trans(_sequence);

  _configDialog.readOptions(*(kapp->config()));
  if (_configDialog.exec())
    {
    _configDialog.writeOptions(*(kapp->config()));
    
    // Reload the config, since something changed
    readOptions(*(kapp->config()));
    }

  // Restart the timer if it should be
  if (contextMenu()->isItemChecked(MENU_ENABLED_ID))
    _mousePollTimer.start(POLL_TIME);
  _mousePaused = 0;
  }

void KGesture::slotChangeIconTimeout()
  {
  setPixmap(KGlobal::iconLoader()->loadIcon("kgesture",
					    KIcon::Small));
  }
